<?php
/**
 * SQL
 * @package     Me
 * @copyright   Copyright (C) 2011 MicroEngine Inc.
 */
class Me_Sql
{
    /** 昇順 */
    const ASC = 'asc';
    /** 降順 */
    const DESC = 'desc';
    /** 作成日時 */
    const CREATE_DATE = 'create_date';
    /** 更新日時 */
    const UPDATE_DATE = 'update_date';
    /**
     * PDOインスタンスを格納
     * @var object
     */
    private $pdo_instance;
    /**
     * SQL文
     * @var string
     */
    private $query;
    /**
     * SQL文を行単位に分割して配列に格納
     * @var array
     */
    private $lines;
    /**
     * SQL文の各行の先頭単語を値に持つ配列
     * @var array
     */
    private $lines_index;
    /**
     * プレースホルダ番号
     * @var int
     */
    private $place_holder_no = 1;
    /**
     * バインドする値の配列
     * @var array
     */
    private $bind_values;

    /**
     * コンストラクタ
     */
    public function __construct()
    {
        // PDOインスタンスを取得
        $this->pdo_instance = Me_Db::get_pdo_instance();
    }

    /**
     * SQL文セット
     * @param string $query SQL文
     */
    public function set_query($query)
    {
        // プロパティを初期化
        $this->lines_index = array();
        $this->bind_values = array();
        $this->query = $query;

        // 改行文字(LF)でSQL文を分割
        $this->lines = explode("\n", $query);
        foreach ($this->lines as &$row) {
            // 空白区切で単語単位に分割
            $row = preg_split('/\s+/', trim($row));
            // インデックスに先頭単語を追加
            $this->lines_index[] = $row[0];
        }
    }

    /**
     * SQL文に値を割り当てる
     * @param mixed(string|int) $line_index SQL文の行番号、または行頭文字列
     * @param mixed(string|int) $word_index 割り当てられる文字列番号、または文字列
     * @param mixed $value 割り当てる値
     * @param mixed(string|int) $remove_start 値がない場合の削除開始位置の文字列番号、または文字列
     * @param mixed(string|int) $remove_end  値がない場合の削除終了位置の文字列番号、または文字列
     */
    public function bind($line_index, $word_index, $value, $remove_start = null, $remove_end = null)
    {
        $this->bind_or_replace($line_index, $word_index, $value, $remove_start, $remove_end, false);
    }

    /**
     * SQL文に値をそのまま置き換える
     * 入力値のチェックはされないのでユーザー入力値をそのまま適用してはいけない。
     * @param mixed(string|int) $line_index SQL文の行番号、または行頭文字列
     * @param mixed(string|int) $word_index 割り当てられる文字列番号、または文字列
     * @param mixed $value 置き換える値
     * @param mixed(string|int) $remove_start 値がない場合の削除開始位置の文字列番号、または文字列
     * @param mixed(string|int) $remove_end  値がない場合の削除終了位置の文字列番号、または文字列
     */
    public function replace($line_index, $word_index, $value, $remove_start = null, $remove_end = null)
    {
        $this->bind_or_replace($line_index, $word_index, $value, $remove_start, $remove_end, true);
    }

    /**
     * bindもしくは、replaceを実行
     * @param mixed(string|int) $line_index SQL文の行番号、または行頭文字列
     * @param mixed(string|int) $word_index 割り当てられる文字列番号、または文字列
     * @param mixed(string|array) $value 割り当てる値もしくは、置き換える値
     * @param mixed(string|int) $remove_start 値がない場合の削除開始位置の文字列番号、または文字列
     * @param mixed(string|int) $remove_end  値がない場合の削除終了位置の文字列番号、または文字列
     * @param boolean $replace_flag bindせずに置き換える場合にtrueを指定する。
     */
    private function bind_or_replace($line_index, $word_index, $value, $remove_start = null, $remove_end = null, $replace_flag = false)
    {
        // 行インデックスを添字番号に変換
        $this->index_check($line_index, $this->lines_index);
        // 行を選択
        $line = &$this->lines[$line_index];
        // 単語インデックスを添字番号に変換
        $this->index_check($word_index, $line);

        if ($value === null) {
            if ($remove_start === null || $remove_end === null) {
                // $valueがnullで、remove指定もされていない場合は、エラーとする。
                throw new Me_Exception(202);
            }
            // 値が無い場合は、$remove_start,$remove_endをみて処理する
            $this->remove_($line_index, $remove_start, $remove_end, false);
        } else {
            // 値がある場合
            if ($replace_flag) {
                // 置換指定の場合は、値をそのまま置き換える。
                $line[$word_index] = $value;
            } else if (is_array($value)) {
                // $valueが配列ならカンマ区切りでプレースホルダを作成する。
                $place_holder_list = array();
                foreach ($value as $val) {
                    $place_holder = ':' . $this->place_holder_no++;
                    $place_holder_list[] = $place_holder;
                    // バインドする値を保管
                    $this->bind_values[$place_holder] = $val;
                }
                $line[$word_index] = implode(', ', $place_holder_list);
            } else {
                // プレースホルダに置換して、値を保管する
                $place_holder = ':' . $this->place_holder_no++;
                $line[$word_index] = $place_holder;
                // バインドする値を保管
                $this->bind_values[$place_holder] = $value;
            }
        }
    }

    /**
     * 指定行もしくは単語を削除する
     * @param mixed(string|int) $line_index SQL文の行番号、または行頭文字列
     * @param mixed(string|int) $remove_start 削除開始位置の文字列番号、または文字列
     * @param mixed(string|int) $remove_end  削除終了位置の文字列番号、または文字列
     */
    public function remove($line_index, $remove_start = null, $remove_end = null)
    {
        $this->remove_($line_index, $remove_start, $remove_end, true);
    }

    /**
     * 指定行もしくは単語を削除する
     * @param mixed(string|int) $line_index SQL文の行番号、または行頭文字列
     * @param mixed(string|int) $remove_start 削除開始位置の文字列番号、または文字列
     * @param mixed(string|int) $remove_end  削除終了位置の文字列番号、または文字列
     * @param boolean $line_remove_flag 行ごと削除する
     */
    private function remove_($line_index, $remove_start = null, $remove_end = null, $line_remove_flag = true)
    {
        // 行インデックスを添字番号に変換
        $this->index_check($line_index, $this->lines_index);
        // 行を選択
        $line = &$this->lines[$line_index];

        $start_flag = ($remove_start === null || $remove_start === '') ? false : true;
        $end_flag = ($remove_end === null || $remove_end === '') ? false : true;
        if ($start_flag && $end_flag) {
            // 削除開始、終了が指定されている場合、開始位置から終了位置までを空にする。
            $this->index_check($remove_start, $line);
            $this->index_check($remove_end, $line);
            for ($i = $remove_start; $i <= $remove_end; $i++) {
                $line[$i] = '';
            }
        } elseif ($start_flag) {
            // 削除開始のみ指定されている場合、指定要素のみ空にする。
            $this->index_check($remove_start, $line);
            $line[$remove_start] = '';
        } elseif ($line_remove_flag) {
            // 行のみ指定のばあいは、全要素を空にする。
            foreach ($line as &$word) {
                $word = '';
            }
        }
    }

    /**
     * 添字チェック
     * 値で指定されている場合は、要素番号を代入する
     * 添字が見つからない場合はエラーとする。
     * @param mixed(string|int) $index
     * @param mixed(string|int) $array
     */
    private function index_check(&$index, &$array)
    {
        if (is_int($index)) {
            if (!isset($array[$index])) {
                // 添字が見つからない場合はエラーとする
                throw new Me_Exception(201);
            }
        } else {
            $index = array_search($index, $array);
            if ($index === false) {
                // 添字が見つからない場合はエラーとする
                throw new Me_Exception(201);
            }
        }
    }

    /**
     * Queryを実行する
     * @return array $rows Query実行結果の配列の配列
     */
    public function exec()
    {
        try {
            // SQL文を組み立てる
            $query = $this->get_query();
            $stmt = $this->pdo_instance->prepare($query);
            // 値をセットする
            foreach ($this->bind_values as $key => $value) {
                $stmt->bindValue($key, $value);
            }
            // クエリ実行
            $stmt->execute();
            $rows = $stmt->fetchAll(PDO::FETCH_ASSOC);
        } catch (PDOException $e) {
            throw new Me_Exception(200, array($e->getMessage()));
            exit;
        } catch (Exception $e) {
            throw new Me_Exception(200, array($e->getMessage()));
            exit;
        }

        return $rows;
    }

    /**
     * SQL文取得
     * @return string $query SQL文
     */
    private function get_query()
    {
        $query = '';
        foreach ($this->lines as $line) {
            foreach ($line as $key => $val) {
                if ($val === '') {
                    unset($line[$key]);
                }
            }
            if (count($line) > 0) {
                $query .= implode(' ', $line) . " \n";
            }
        }
        return $query;
    }

    /**
     * SQL文とその分割状態を表示する
     */
    public function debug()
    {
        var_dump($this->query);
        var_dump($this->lines);
        var_dump($this->lines_index);
    }
}